/* CountVariants (Ergaenzung zu Element)
   Lutz Prechelt
   1997-01-02
   RCS: $Id: CountVariants.java,v 1.1 1997/01/22 16:43:06 prechelt Exp prechelt $
*/

import java.util.Enumeration;

/** Berechne die Anzahl von Varianten in einem Element
*/
class CountVariants implements ElementAction {
  int variants = 0;  // Anzahl Varianten im besuchten Element

  CountVariants(Element e) {
     e.perform(this);
  }

  /** Berechnet die Anzahl von Varianten in einem StringElement, naemlich 1.
  */
  public void stringAction(StringElement e) {
     variants = 1;
  }
  
  /** Berechnet die Anzahl von Varianten in einem AndElement:
      Produkt der Variantenzahlen der enthaltenen Elemente (unter Auslassung
      von leeren Elementen)
  */
  public void andAction(AndElement u) {
     int allVariants = 1;
     Enumeration n = u.elems.elements();
     if (!n.hasMoreElements()) {  // leeres AndElement hat nur 0 Varianten,
        variants = 0;             // nicht 1!
        return;
     }
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        e.perform(this);    // berechne variants fuer ein Unterelement
        if (variants > 0)   // ignoriere leere AndElemente!
           allVariants *= variants;  // multipliziere andernfalls Varianten
     }
     variants = allVariants;
  }
  
  
  /** Berechnet die Anzahl von Varianten in einem OrElement:
      Summe der Variantenanzahlen der enthaltenen Elemente.
  */
  public void orAction(OrElement o) {
     int allVariants = 0;
     Enumeration n = o.elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        e.perform(this);          // berechne variants fuer ein Unterelement
        allVariants += variants;  // und kumuliere diese Werte
     }
     variants = allVariants;  // Liefere Resultat
  }


  /** Hauptprogramm zum Testen
  */
  public static void main (String args[]) {
     AndElement u = new AndElement();
     OrElement modal = new OrElement();
     OrElement verb  = new OrElement();
     u.add(new StringElement("Wer "));
     modal.add(new StringElement("kann "));
     modal.add(new StringElement("will "));
     u.add(modal);
     u.add(new StringElement("denn schon "));
     verb.add(new StringElement("denken "));
     verb.add(new StringElement("handeln "));
     u.add(verb);
     u.add(new StringElement("?"));
     CountVariants c = new CountVariants(u);
     System.out.println("u has " + c.variants + " Variants");
  }
}