/* Tupel
   #Mm1744   
   Lutz Prechelt, Barbara Unger
   1996-12-05
   RCS: $Id: TupelMain.java,v 1.2 1997/01/22 16:42:12 prechelt Exp prechelt $
*/
import java.awt.*;
import java.util.Enumeration;
import java.util.Vector;

/*
Dieses Programm verwaltet Mengen von Tupeln.
Ein Tupel besteht aus mehreren Feldern, z.B. Name, Vorname, Telefonnummer.
Es gibt potentiell mehrere solche Tupeltypen und zu jedem davon einen 
Tupelmengentyp.
Das Programm enthaelt ferner GUI-Klassen, mit denen Tupel eines
bestimmten Tupeltyps eingelesen werden koennen und mit denen Tupelmengen
eines bestimmten Tupeltyps angezeigt werden koennen.
Derzeit wird nur der Tupeltyp NTTupel voll unterstuetzt.

Hier die Klassenhierarchie:
  interface Tupel
     class NTTupel
     class NATupel
  class Tupelmenge
  class java.awt.Frame
     class NTEinleser
     class NAEinleser
     abstract class TupelAnzeiger
        class NTTupelAnz1
        abstract class TupelAnzA
           class NTTupelAnz2
     class TupelMain
#M
#M*** ENTWURFSMUSTER: ***
#M
#MTupelAnzeiger ist die abstrakte Oberklasse fuer **Beobachter**
#Mder Datenstruktur Tupelmenge.
#M
#MTupelAnzA ist die abstrakte Oberklasse fuer eine Gruppe solcher
#MBeobachter, die sich hinsichtlich Auswahl, Anordnung und Formatierung
#Mder darzustellenden Tupel mit Hilfe einer **Schablonenmethode**
#Mvariieren lassen.
*/


//------------------------ Tupel ------------------------------
/**
  Gemeinsame Schnittstelle aller Tupeltypen.
*/
interface Tupel {
  /** Liest ein Tupel ein und liefert es bei der angegebenen Tupelmenge
      ab.
  */
  public void getTupel(Tupelmenge m);
}

//------------------------ NTTupel ------------------------------
/**
  * Eine Implementierung von Tupel fuer Namen und Telefonnummern
  */
class NTTupel implements Tupel {
  private String name;      // Name einer Person
  private String vorname;   // Vorname der Person
  private String telefon;   // Telefonnummer der Person

  public String getName() { return name; }

  public String getVorname() { return vorname; }

  public String getTelefon() { return telefon; }

  public void setName(String n) { name = n; }

  public void setVorname(String v) { vorname = v; }
  
  public void setTelefon(String t) { telefon = t; }
  
  /** Liest ein Tupel ein. Die genaue Strategie hierfuer ist
      in getTupel verkapselt. getTupel garantiert nur,
      dass das Tupel nach dem Einlesen bei der Tupelmenge m gemeldet wird
      und danach auch selbst die richtigen Feldwerte enthaelt.
      getTupel garantiert NICHT, dass das Einlesen bei Rueckkehr aus
      getTupel bereits abgeschlossen ist.
  */
  public void getTupel(Tupelmenge m) {
     NTEinleser e = new NTEinleser(this, m); // einlesen mit GUI, asynchron
  }
}


//------------------------ NTEinleser ------------------------------
/**
    GUI-Einleseklasse fuer NTTupel.
    Erst hierin wird endgueltig die Entscheidung getroffen, dass das
    Einlesen mit einem GUI erfolgt.
*/
final class NTEinleser extends Frame {
  private Button     ok;
  private TextField  t1, t2, t3;
  private Label      l1, l2, l3;
  private NTTupel    tupel;
  private Tupelmenge m;

  /** Erzeugt Fenster mit drei Textfeldern und einem OK-Knopf zum
      Einlesen der Teilwerte eines NTTupel
  */
  public NTEinleser(NTTupel t, Tupelmenge m) {
     super("Tupel eingeben");
     Panel p = new Panel();
     p.setLayout(new GridLayout(3, 2));
     this.setLayout(new FlowLayout());
     l1 = new Label("Vorname");
     l2 = new Label("Name");
     l3 = new Label("Telefon");
     t1 = new TextField();
     t2 = new TextField();
     t3 = new TextField();
     ok = new Button("O.K.");
     p.add(l1);  p.add(t1);  // trage Label/Text-Paare in 3 Zeilen in p ein
     p.add(l2);  p.add(t2);
     p.add(l3);  p.add(t3);
     this.add(p);
     this.add(ok);
     this.tupel = t; // wird gebraucht, wenn OK-Knopf gedrueckt wurde.
     this.m = m;     // dito
     this.pack();    // berechne layout
     this.show();    // zeige Dialogfenster an
  }

  /** Ereignisbehandlungsmethode. Verstaut die eingelesenen Werte im
      Tupel, wenn der OK-Knopf gedrueckt wurde und meldet das Tupel
      an die Tupelmenge.
  */
  public boolean action(Event ev, Object target) {
     if (ev.id == Event.ACTION_EVENT) {
       if (ev.target == ok) {      // OK-Knopf gedrueckt
         tupel.setVorname(t1.getText());
         tupel.setName(t2.getText());
         tupel.setTelefon(t3.getText());
         m.neuesTupel(tupel);
         this.hide();              // entferne Frame von Bildschirm
         this.dispose();           // gib seine Resourcen frei
         return (true);
       }
       else {
         return (false);
       }
     }
     else {
       return (false);
     }
  }
}


//------------------------ NATupel ------------------------------
/**
    Eine Implementierung von Tupel fuer Namen und Adressen
*/
class NATupel implements Tupel {
  private String name;      // Name einer Person
  private String vorname;   // Vorname der Person
  private String strasse;   // Strasse und Hausnummer der Person
  private String ort;       // Postleitzahl und Ort der Person

  public String getName() { return name; }

  public String getVorname() { return vorname; }

  public String getStrasse() { return strasse; }

  public String getOrt() { return ort; }

  public void setName(String n) { name = n; }

  public void setVorname(String v) { vorname = v; }
  
  public void setStrasse(String s) { strasse = s; }

  public void setOrt(String o) { ort = o; }
  
  /** Liest ein Tupel ein. Die genaue Strategie hierfuer ist
      in getTupel verkapselt. getTupel garantiert nur,
      dass das Tupel nach dem Einlesen bei der Tupelmenge m gemeldet wird
      und danach auch selbst die richtigen Feldwerte enthaelt.
      getTupel garantiert NICHT, dass das Einlesen bei Rueckkehr aus
      getTupel bereits abgeschlossen ist.
  */
  public void getTupel(Tupelmenge m) {
     NAEinleser e = new NAEinleser(this, m); // einlesen mit GUI, asynchron
  }
}


//------------------------ NAEinleser ------------------------------
/**
    GUI-Einleseklasse fuer NATupel.
    Erst hierin wird also endgueltig die Entscheidung getroffen, dass das
    Einlesen mit einem GUI erfolgt.
*/
final class NAEinleser extends Frame {
  private Button     ok;
  private TextField  t1, t2, t3, t4;
  private Label      l1, l2, l3, l4;
  private NATupel    tupel;
  private Tupelmenge m;

  /** Erzeugt Fenster mit drei Textfeldern und einem OK-Knopf zum
      Einlesen der Teilwerte eines NTTupel
  */
  public NAEinleser(NATupel t, Tupelmenge m) {
     super("Tupel eingeben");
     Panel p = new Panel();
     p.setLayout(new GridLayout(4, 2));
     this.setLayout(new FlowLayout());
     l1 = new Label("Vorname");
     l2 = new Label("Name");
     l3 = new Label("Strasse");
     l4 = new Label("Ort");
     t1 = new TextField();
     t2 = new TextField();
     t3 = new TextField();
     t4 = new TextField();
     ok = new Button("O.K.");
     p.add(l1);  p.add(t1);  // trage Label/Text-Paare in p ein
     p.add(l2);  p.add(t2);
     p.add(l3);  p.add(t3);
     p.add(l4);  p.add(t4);
     this.add(p);
     this.add(ok);
     this.tupel = t; // wird gebraucht, wenn OK-Knopf gedrueckt wurde.
     this.m = m;     // dito
     this.pack();    // berechne layout
     this.show();    // zeige Dialogfenster an
  }

  /** Ereignisbehandlungsmethode. Verstaut die eingelesenen Werte im
      Tupel, wenn der OK-Knopf gedrueckt wurde und meldet das Tupel
      an die Tupelmenge.
  */
  public boolean action(Event ev, Object target) {
     if (ev.id == Event.ACTION_EVENT) {
       if (ev.target == ok) {      // OK-Knopf gedrueckt
         tupel.setVorname(t1.getText());
         tupel.setName(t2.getText());
         tupel.setStrasse(t3.getText());
         tupel.setOrt(t4.getText());
         m.neuesTupel(tupel);
         this.hide();              // entferne Frame von Bildschirm
         this.dispose();           // gib seine Resourcen frei
         return (true);
       }
       else {
         return (false);
       }
     }
     else {
       return (false);
     }
  }
}


//------------------------ Tupelmenge ------------------------------
/** 
    Verwaltet einen ganzen Haufen von Tupeln und erlaubt, einzelne Tupel
    diesem Haufen zuzufuegen.
    Verwaltet ferner einen Haufen von Tupelanzeigern, die
    aufgerufen werden, wenn ein Tupel zugefuegt wird.
*/
class Tupelmenge {
  private Vector tupelanzeigervektor = new Vector();  // eingetragene Anzeiger
  private Vector tupelvektor = new Vector();          // die Tupel
    
  /** wird von einem Tupelerzeuger aufgerufen um ein neues Tupel in die
      Tupelmenge einzubringen
  */
  synchronized void neuesTupel(Tupel t) {
     tupelvektor.addElement((Object)t);
     Enumeration e = tupelanzeigervektor.elements();
     while (e.hasMoreElements())
        ((TupelAnzeiger)e.nextElement()).neuesTupel(t);
  }

  void neuerAnzeiger(TupelAnzeiger ta){
    tupelanzeigervektor.addElement((Object)ta);
    ta.neueTupel(tupelvektor);
  }
}


//------------------------ TupelAnzeiger ------------------------------
/**
    TupelAnzeiger stellt mehrere Tupel in einem eigenen Fenster dar.
    Jede Unterklasse legt eine eigene Darstellungsform fest.
    TupelAnzeiger garantieren, dass sie keine Veraenderungen auf den
    Tupel-Objekten durchfuehren (so dass man ihnen 'Originale' anvertrauen
    kann und nicht dauernd Kopien machen muss).
    Ein Programm kann eine ganze Reihe von TupelAnzeiger-Objekten herstellen
    und ihnen die gleichen Tupel uebergeben.
*/
abstract class TupelAnzeiger extends Frame {
  private Vector tupelvektor = new Vector();
  
  TupelAnzeiger(String name) {
     super(name);
  }
     
  /** Wird vom Benutzer des Tupelanzeigers aufgerufen, um zu verkuenden,
      dass ein neues Tupel zu den bislang schon dargestellten hinzukommen
      soll.
  */
  abstract synchronized void neuesTupel(Tupel t);

  /** wird vom Benutzer des Tupelanzeigers aufgerufen, um zu verkuenden,
      dass mehrere neue Tupel zu den bislang schon dargestellten hinzukommen
      sollen.
      Der Aufrufer garantiert, dass sich im Vector ausschliesslich 
      'Tupel'-Objekte befinden.
      neueTupel() garantiert, dass es 'neuetupel' nicht aendert.
  */
  abstract synchronized void neueTupel(Vector neuetupel);
}


//------------------------ NTTupelAnz1 ------------------------------
/**
    Stellt NTTupel-Objekte in der Reihenfolge ihrer Anlieferung
    mit allen Einzelteilen in einem einfachen Format dar.
    Die Benutzung von Tupelobjekten anderer Tupeltypen fuehrt zu
    ClassCastException.
*/
final class NTTupelAnz1 extends TupelAnzeiger {
  private TextArea textarea;

  /** oeffnet neues Fenster mit einem TextArea darin, in dem spaeter die
      Tupel angezeigt werden sollen
  */
  NTTupelAnz1(String name) {
     super(name);
     this.setLayout(new FlowLayout());
     textarea = new TextArea(5, 15);
     this.add(textarea);  // textarea ist einziges Element im Frame
     this.pack();
     this.show();
  }

  synchronized void neuesTupel(Tupel t) {
     NTTupel nt = (NTTupel)t;
     // neues Tupel einfach hinten an bisherige Darstellung anhaengen:
     textarea.appendText(nt.getVorname() + " " +
                         nt.getName()    + ", " +
                         nt.getTelefon() + "\n");
  }

  synchronized void neueTupel(Vector neuetupel) {
     Enumeration e = neuetupel.elements();
     while (e.hasMoreElements())
        neuesTupel((Tupel)e.nextElement());
  }
}


//------------------------ TupelAnzA ------------------------------
/**
    TupelAnzA stellt Tupelmengen in einer Form dar, die durch Einsetzen
    einer Selektiermethode (Tupel ganz zurueckweisen), 
    einer Vergleichsmethode (Tupelreihenfolge bestimmen) und
    einer Formatiermethode (Tupelfelder beliebig auswaehlen und formatieren)
    an viele Zwecke angepasst werden kann.
*/
abstract class TupelAnzA extends TupelAnzeiger {
  /** tupelvektor enthaelt alle darzustellenden Tupel (nach dem Ausfiltern!),
      die Reihenfolge spiegelt die Darstellungsreihenfolge wider.
      tupelstringvektor enthaelt parallel zu tupelvektor die endgueltigen
      Anzeigeformen der Tupel.
      Die Inhalte von tupelstringvektor werden direkt in textarea geschrieben.
      textarea enthaelt die endgueltige Gesamtdarstellung.
  */
  private Vector tupelvektor;       // darzustellende Tupel
  private Vector tupelstringvektor; // dazu korrespondierend: ihre Darstellung
  private TextArea textarea;

  TupelAnzA(String name) {
     super(name);
     tupelvektor = new Vector();
     tupelstringvektor = new Vector();
     this.setLayout(new FlowLayout());
     textarea = new TextArea(4, 16);
     this.add(textarea);  // textarea ist einziges Element im Frame
     this.pack();
     this.show();
  }

  /** Liefert die gewuenschte Darstellung von Tupel a als String.
      Der String darf Null, eines oder mehrere Newline-Zeichen enthalten.
  */
  abstract String formatiere(Tupel a);

  /** Liefert true, wenn das Tupel dargestellt werden soll und 
              false wenn nicht
  */
  abstract boolean selektiere(Tupel a);

  /** Liefert true, wenn Tupel a vor Tupel b dargestellt werden soll und
              false, wenn Tupel b vor Tupel a dargestellt werden soll.
  */
  abstract boolean vergleiche(Tupel a, Tupel b);

  /** realisiert das Zufuegen eines neuen Tupels.
      Es wird zunaechst mittels selektiere() bestimmt, ob das Tupel ueberhaupt
      zugefuegt werden soll, dann wird es ggf. in der Methode sortierenEin()
      mittels vergleiche an die richtige Stelle in der Darstellung
      gebracht und mittels formatiere() in einen String des gewuenschten
      Darstellungsformats verwandelt.
#M      *** ENTWURFSMUSTER: ***
#M      neuesTupel() ist zusammen mit seiner Hilfsmethode sortiereEin() eine
#M      **Schablonenmethode**, wobei die in den Unterklassen aufgefuellten
#M      Luecken aus den Methoden selektiere(), formatiere() und vergleiche()
#M      bestehen.
  */
  synchronized void neuesTupel(Tupel t) {
     Tupel nt = (Tupel)t;
     if (!selektiere(nt))
        return;   // Dieses Tupel soll nicht angezeigt werden
     sortiereEin(nt);  // tut nt in tupelvektor und tupelstringvektor
     stelleDar();
  }

  /** Analog zu neuesTupel(), jedoch in einer Schleife fuer mehrere Tupel
      zugleich, wobei die Darstellung nur einmal am Ende aufgefrischt wird.
  */
  synchronized void neueTupel(Vector neuetupel) {
     Enumeration e = neuetupel.elements();
     while (e.hasMoreElements()) {
        Tupel nt = (Tupel)e.nextElement();
        if (selektiere(nt))
           sortiereEin(nt);
        else
           ; // ignoriere Tupel
     }
     stelleDar();
  }

  /** Sucht die Stelle im tupelvektor, an die das neue Tupel gehoert
      und fuegt es dort ein. Erzeugt ausserdem die endgueltige Anzeigeform
      fuer das Tupel und fuegt diese in tupelstringvektor ein.
  */
  private void sortiereEin(Tupel nt) {
    int wohin = 0;  // index, bei dem neues Tupel einzusortieren ist.
    while(wohin < tupelvektor.size() &&
          vergleiche((Tupel)tupelvektor.elementAt(wohin), nt))
      wohin++;
    // nun zeigt 'wohin' die Einfuegeposition an.
    // verschiebe nun die restlichen Elemente um eins nach hinten
    // und zwar sowohl in tupelvektor als auch im tupelstringvektor!
    int sz = tupelvektor.size(),
        k  = sz-1;  // hinten anfangen
    tupelvektor.setSize(sz+1);
    tupelstringvektor.setSize(sz+1);
    while (k >= wohin) {
       tupelvektor.setElementAt(tupelvektor.elementAt(k), k+1);
       tupelstringvektor.setElementAt(tupelstringvektor.elementAt(k), k+1);
       k--;
    }
    // nun fuege das neue Tupel in tupelvektor und seine fertig
    // formatierte Form in tupelstringvektor an Position 'wohin' ein:
    tupelvektor.setElementAt(nt, wohin);
    tupelstringvektor.setElementAt(formatiere(nt), wohin);
  }

  /** Stellt den Inhalt von tupelstringvektor neu in textarea dar.
  */
  private void stelleDar() {
     StringBuffer sb = new StringBuffer();
     Enumeration e = tupelstringvektor.elements();
     sb.ensureCapacity(textarea.getText().length() + 100); // optimiert!!!
     while (e.hasMoreElements())
        sb.append((String)e.nextElement());
     textarea.setText(sb.toString());  // komplett neuer Text fuer textarea
  }
}


//------------------------ NTTupelAnz2 ------------------------------
/**
    NTTupelAnz2 stellt NTTupel dar, wobei
    1. Tupel, deren Telefonnummer leer ist, weggelassen werden und
    2. die Tupel nach Nachname sortiert werden
    Die Benutzung von Tupelobjekten anderer Tupeltypen fuehrt zu
    ClassCastException.
#M    *** ENTWURFSMUSTER: ***
#M    NTTupelAnz2 vervollstaendigt die **Schablonenmethode** neuesTupel()
#M    von TupelAnzA
*/
final class NTTupelAnz2 extends TupelAnzA {

  NTTupelAnz2(String name) {
     super(name);
  }

  /** Liefert die gewuenschte Darstellung von Tupel a als einzeiligen String
  */
  String formatiere(Tupel a) {
     NTTupel b = (NTTupel)a;
     return (b.getVorname() + " " +
             b.getName()    + ", Tel.: " +
             b.getTelefon() + "\n");
  }

  /** Tupel soll nur dargestellt werden, wenn es eine Telefonnummer enthaelt.
  */
  boolean selektiere(Tupel a) {
     return !((NTTupel)a).getTelefon().equals("");
  }

  /** a soll vor b dargestellt werden, wenn es einen alphabetisch kleineren
      Nachnamen hat.
  */
  boolean vergleiche(Tupel a, Tupel b) {
     return (((NTTupel)a).getName().compareTo(((NTTupel)b).getName()) < 0);
  }
}


//------------------------ TupelMain ------------------------------
/**
    Hauptprogramm. Erzeugt ein Hauptfenster mit zwei Knoepfen, eine
    Tupelmenge und zwei Tupelanzeiger. Einer der Knoepfe erzeugt jeweils
    ein NTTupel und fuegt es der Tupelmenge zu.
    Es gibt keinen statischen Typschutz zwischen dem tatsaechlichen Tupeltyp
    in der Tupelmenge und dem Tupeltyp, der von den Tupelanzeigern erwartet
    wird.
#M    *** ENTWURFSMUSTER: ***
#M    Die beiden Tupelanzeiger werden als Beobachter bei der Tupelmenge 
#M    eingetragen.
*/
public final class TupelMain extends Frame {
   Button     quit = new Button("Quit");
   Button     get  = new Button("mehr!");
   Tupelmenge alleTupel = new Tupelmenge();

   public TupelMain(String name) {
      super(name);
      this.setLayout(new FlowLayout());
      this.add(quit);
      this.add(get);
      this.pack();
      this.show();
   }   

   /** Ereignisbehandlung fuer das gesamte Programm.
   */
   public boolean action(Event ev, Object target) {
     if (ev.id == Event.ACTION_EVENT) {
       if (ev.target == quit) {  // Quit-Knopf gedrueckt
          this.hide();       // mache Hauptfenster unsichtbar
          System.exit(0);    // Verlasse das Programm
       }
       else if (ev.target == get) { // "mehr!"-Knopf gedrueckt
          Tupel t = new NTTupel();
          t.getTupel(alleTupel);
       }
       else {
         System.out.println ("ACTION_EVENT: " + ev.arg.toString()); //!!!
         return (false);
       }
     }
     else {
       System.out.println ("Event: " + ev.arg.toString() + " " + ev.id); //!!!
       return (false);
     }
     return (true);  // all but the 'return false;' cases have been handled
   }    

   /* ------------ m a i n : ------------ */
   public static void main(String args[]) {
      TupelMain mainwindow = new TupelMain("TupelGucker");
      TupelAnzeiger anz1 = new NTTupelAnz1("chronologisch");
      TupelAnzeiger anz2 = new NTTupelAnz2("nach Nachname");
      mainwindow.alleTupel.neuerAnzeiger(anz1);
      mainwindow.alleTupel.neuerAnzeiger(anz2);
   }
}
